#include <png.h>

#include "bitmap_read.h"

/* read PNG into BITMAP using libpng */
int bitmap_read_png(BITMAP **bmp, FILE *f) {
  int state = BR_OK;
  unsigned char hdr[8];
  /* tiny check first :) */
  if (bmp == NULL)
    return BR_BAD_BITMAP;
  if (f == NULL)
    return BR_BAD_FILE;
  /* check file type */
  if (fread(hdr, 1, sizeof(hdr), f) < sizeof(hdr)) return BR_INVALID_FORMAT;
  if (png_sig_cmp(hdr, 0, sizeof(hdr))) return BR_INVALID_FORMAT;
  /* preparing read of PNG */
  png_structp png_ptr = png_create_read_struct(
    PNG_LIBPNG_VER_STRING, (png_voidp) NULL, NULL, NULL);
  png_infop info_ptr = NULL, end_info = NULL;
  if (png_ptr != NULL) {
    /* some auxiliary structures ... */
    info_ptr = png_create_info_struct(png_ptr);
    end_info = png_create_info_struct(png_ptr);
    if (info_ptr != NULL && end_info != NULL) {
      unsigned long int width, height;
      int depth, color, interlace, comp, filter;
      /* libpng will read from f */
      png_init_io(png_ptr, f);
      png_set_sig_bytes(png_ptr, sizeof(hdr)); /* already checked */
      /* get PNG information */
      png_read_info(png_ptr, info_ptr);
      /* get all what we need from PNG headers */
      png_get_IHDR(png_ptr, info_ptr, &width, &height, &depth, &color,
        &interlace, &comp, &filter);
      /* compute actual bit count per pixel */
      switch (color) {
        case PNG_COLOR_TYPE_RGB: depth *= 3; break;
        case PNG_COLOR_TYPE_RGB_ALPHA: depth *= 4; break;
      }
      if ((*bmp = createbitmap2(width, height, depth, 4)) != NULL) {
/*
#if __BYTE_ORDER == __BIG_ENDIAN
      png_set_bgr(png_ptr);
#endif
*/
        /* read image at once */  
        png_read_image(png_ptr, (png_bytepp) (*bmp)->lines);
        /* finish it */
        png_read_end(png_ptr, end_info);
      } else state = BR_OUT_OF_MEMORY;
    } else state = BR_AUX_STRUCT_ERROR;
  } else state = BR_AUX_STRUCT_ERROR;
  /* klid */
  png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
  return state; /* konec! */
}
