#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "api.h"
#include "stack.h"

/*
 *  Alokuje pamet more (sea.array).
 *
 *  Param:
 *      SEA sea: more (musi mit nastavene sea.xs, sea.ys)
 *  
 *  Navratova hodnota:
 *      0 uspech
 *     -1 neuspech
 */
int sea_malloc(SEA *sea) {
  if (sea == NULL || sea->xs <= 0 || sea->ys <= 0) return -1;
  if ((sea->array = (SEAFIELD **) malloc(sea->ys*sizeof(SEAFIELD *)
    + sea->xs*sea->ys*sizeof(SEAFIELD))) == NULL)
    return -1;
  *sea->array = (SEAFIELD *) &sea->array[sea->ys];
  for (int i = 1; i < sea->ys; i++)
    sea->array[i] = &(*sea->array)[i*sea->xs];
  memset(*sea->array, 0, sea->xs*sea->ys*sizeof(SEAFIELD));
  return 0;
}

/*
 *  Uvolni pamet alokovanou pro more.
 *
 *  Param:
 *      SEA sea: more (musi mit nastavene sea.xs, sea.ys)
 */
void sea_free(SEA *sea) {
  if (sea == NULL || sea->xs <= 0 || sea->ys <= 0) return;
  if (sea->array != NULL) free(sea->array);
}

/*
 *  Naplni vsechna policka more hodnotou fill
 *
 *  Param:
 *      SEA sea: more
 *      SEAFIELD fill: hodnota na naplneni more
 */
void sea_fill(SEA sea, SEAFIELD fill) {
  for (int i = 0; i < sea.xs*sea.ys; i++) (*sea.array)[i] = fill;
}

/*
 *  Rozhazi do more miny podle nastavene sea.mines.
 */
void sea_randomize(SEA sea) {
  int j;
  for (int i = 0; i < sea.mines; ) {
    j = rand() % (sea.xs*sea.ys);
    if (!(*sea.array)[j].mine) {
      (*sea.array)[j].mine = 1;
      i++;
    }
  }
}

/*
 *  Na kazdem policku ulozi do count, kolik je kolem min.
 */
void sea_count_mines(SEA sea) {
  int i, j, k, l;
  for (i = 0; i < sea.xs; i++)
    for (j = 0; j < sea.ys; j++)
      for (k = i > 0 ? i - 1 : 0; k <= i + 1 && k < sea.xs; k++)
        for (l = j > 0 ? j - 1 : 0; l <= j + 1 && l < sea.ys; l++) {
          if ((k!=i || l!=j) && sea.array[l][k].mine) sea.array[j][i].count++;
        }
}

/*
 *  Nacte miny do more ze souboru. Soubor je tvoren znaky '0' a '1', kde 1 je
 *  mina, 0 je voda. Predpoklada platne hodnoty sea.xs, sea.ys. Nastavi
 *  hodnotu sea.mines, ale jen v pripade, ze nacitani probehne uspesne!!!
 *
 *  Param:
 *      SEA sea: more
 *      char *file: cesta souboru, ze ktereho se nacita
 *
 *  Navratova hodnota:
 *      0 uspech
 *     -1 neuspech
 */
int sea_load(SEA *sea, char *file) {
  FILE *f = fopen(file, "r");
  if (f == NULL) return -1;
  int i = 0, c, count = 0;
  while (!feof(f) && i < sea->xs*sea->ys)
    if ((c = fgetc(f)) == '0' || c == '1') {
      (*sea->array)[i++].mine = c == '1';
      if (c == '1') count++;
    }
  fclose(f);
  return (i != sea->xs*sea->ys) ? -1 : (sea->mines = count), 0;
}

/*
 *  Nastavi vsechna policka v mori jako schovana.
 */
void sea_hide(SEA sea) {
  for (int i = 0; i < sea.xs*sea.ys; i++) (*sea.array)[i].hidden = 1;
}

/*
 *  Oznaci policko na souradnicich. U oznaceneho policka oznaceni zrusi.
 *
 *  Param:
 *      SEA sea: more
 *      x, y: souradnice oznaceni
 *
 *  Navratova hodnota:
 *     -3 souradnice jsou mimo hranice more
 *     -1 policko je uz odkryte
 *      0 v poradku
 */
int sea_flag(SEA sea, int x, int y) {
  if (x < 0 || y < 0 || x >= sea.xs || y >= sea.ys) return -3;
  if (sea.array[y][x].hidden) {
    sea.array[y][x].flag ^= 1;
    return 0;
  }
  return -1;
}

/*
 *  Odkryje policko na souradnicich.
 *
 *  Param:
 *      SEA sea: more
 *      x, y: souradnice odkryti
 *
 *  Navratova hodnota:
 *     -3 souradnice jsou mimo hranice more
 *     -2 policko je oznacene
 *     -1 policko je uz odkryte
 *      0 v poradku
 *      1 mina
 *      2 vitezstvi
 */
int sea_uncover(SEA sea, int x, int y) {
  if (x < 0 || y < 0 || x >= sea.xs || y >= sea.ys) return -3;
  if (sea.array[y][x].flag) return -2;
  if (!sea.array[y][x].hidden) return -1;
  if (sea.array[y][x].mine) return 1;
 
  STACK stack;
  stack_init(&stack);
  stack_push(&stack, (XY){x,y});
  XY xy;
  while (!stack_empty(stack)) {
    stack_pop(&stack, &xy);
    if (!sea.array[xy.y][xy.x].hidden || sea.array[xy.y][xy.x].flag) continue;
    sea.array[xy.y][xy.x].hidden = 0;
    if (sea.array[xy.y][xy.x].count) continue;
    for (int k = xy.x > 0 ? xy.x - 1 : 0; k <= xy.x + 1 && k < sea.xs; k++)
      for (int l = xy.y > 0 ? xy.y - 1 : 0; l <= xy.y + 1 && l < sea.ys; l++)
        if (k!=xy.x || l!=xy.y) stack_push(&stack, *&(XY){k,l});
  }
  stack_clear(&stack);
 
  // takto me nkolik proces sdlet toto API a nebudou si pepisovat tae
  int count = sea.xs*sea.ys-sea.mines;
  for (int i = 0; i < sea.xs*sea.ys; i++)
    if (!(*sea.array)[i].hidden) count--;
  if (count == 0) return 2;
  return 0;
}

/*
 *  Odkryje blok policek 3x3 podle souradnic stredu.
 *
 *  Param:
 *      SEA sea: more
 *      x, y: souradnice odkryti
 *
 *  Navratova hodnota:
 *     -3 souradnice jsou mimo hranice more
 *      0 v poradku
 *      1 mina
 *      2 vitezstvi
 */
int sea_uncover_block(SEA sea, int x, int y) {
  int stat, i, j;
  if ((stat = sea_uncover(sea, x, y)) == -3) return stat;
  for (i = x - 1; i <= x + 1; i++) for (j = y - 1; j <= y + 1; j++)
    if ((i != x || j != y) && stat <= 0) stat = sea_uncover(sea, i, j);
  return stat > 0 ? stat : 0;
}

/*
 *  Odkryje pouze miny. (vola se v pripade prohry)
 */
void sea_uncover_mines(SEA sea) {
  for (int i = 0; i < sea.xs*sea.ys; i++)
    if ((*sea.array)[i].mine) (*sea.array)[i].hidden = 0;
}

/*
 *  Oznakuje miny. (vol se pi vhe)
 */
void sea_flag_mines(SEA sea) {
  for (int i = 0; i < sea.xs*sea.ys; i++)
    if ((*sea.array)[i].mine) (*sea.array)[i].flag = 1;
}

/*
 *  Odkryje cele more. (vola se v pripade vitezstvi)
 */
/*void sea_uncover_all(SEA sea) {
  for (int i = 0; i < sea.xs*sea.ys; i++) (*sea.array)[i].hidden = 0;
}*/
