#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <string.h>

#include "api.h"
#include "stack.h"

#define MIN_WIDTH 10
#define MAX_WIDTH 37

#define MIN_HEIGHT 10
#define MAX_HEIGHT 17 
  
/* nakresl mapku */
void sea_print(SEA sea) {
  int i, j;
  SEAFIELD s;
  
  printf("  "); // odsazen
  for (i = 1; i < sea.xs; i++)
    if ((i % 10) == 0)
      printf("%d", i);
    else
      printf("  ");
  printf("\n  ");
  for (i = 1; i <= sea.xs; i++)
    printf("%2d", i % 10);
  printf("\n  /");
  for (i = 0; i < 2*sea.xs; i++) putchar('-');
  printf("\\\n");
  for (j = 0; j < sea.ys; j++) {
    printf("%2d|", j+1);
    for (i = 0; i < sea.xs; i++) {
      s = sea.array[j][i];
      if (s.flag)
        putchar('X');
      else
        if (s.hidden)
	  putchar('.');
	else
	  if (s.mine)
	    putchar('*');
	  else
	    if (s.count > 0)
	      printf("%d", s.count);
	    else
	      putchar(' ');
      putchar(' ');
    }
    printf("|%d\n", j+1);
  }
  printf("  \\");
  for (i = 0; i < 2*sea.xs; i++) putchar('-');
  printf("/\n  ");
  for (i = 1; i <= sea.xs; i++)
    printf("%2d", i % 10);
  printf("\n  ");
  for (i = 1; i < sea.xs; i++)
    if ((i % 10) == 0)
      printf("%d", i);
    else
      printf("  ");
  putchar('\n');
}

char *strip(char *buffer) {
  if (buffer == NULL) return 0;
  unsigned char *p;
  for (p = (unsigned char *) buffer; isspace(*p); p++);
  buffer = (char *) p;
  for (p = (unsigned char *) strrchr(buffer, '\0'); (char *) p != buffer &&
    isspace(*(p-1)); p--);
  *p = '\0';
  return buffer;
}

int main(int argc, char **argv) {
  srand(time(0));

  if (argc < 4) {
    printf("nvod : miny x y n [soubor s polem-moem]\n\n"
           "ka : %u < x < %u\n"
           "Vka : %u < y < %u\n"
           "Poet : 1 < n < x * y - 10\n"
           "Soubor: obsahuje pedem nachystan pole\n\n",
           MIN_WIDTH, MAX_WIDTH, MIN_HEIGHT, MAX_HEIGHT);
    return 1;
  }

  SEA sea = { .mines = -1 };
  char *p;
  sea.xs = strtol(argv[1], &p, 0);
  if (*p != '\0') { printf("patn ka: '%s'\n", argv[1]); return 1; }
  if (sea.xs < MIN_WIDTH || sea.xs > MAX_WIDTH) {
    printf("ka %d je mimo rozsah.\n", sea.xs);
    return 2;
  }
  sea.ys = strtol(argv[2], &p, 0);
  if (*p != '\0') { printf("patn vka: '%s'\n", argv[2]); return 3; }
  if (sea.ys < MIN_HEIGHT || sea.ys > MAX_HEIGHT) {
    printf("Vka %d je mimo rozsah.\n", sea.ys);
    return 4;
  }

  if (sea_malloc(&sea) < 0) {
    printf("Nelze vytvoit moe.\n");
    return 5;
  }
  
  if (argc > 4 && sea_load(&sea, argv[4]) < 0)
    printf("Nepodailo se nast moe ze souboru.\n");
  else if (sea.mines == 0)
    printf("Moe je przdn, a to nechceme.\n");
  if (sea.mines <= 0) {
    sea.mines = strtol(argv[3], &p, 0);
    if (*p != '\0') {
      printf("patn poet min: '%s'\n", argv[3]);
      sea_free(&sea);
      return 6;
    }
    if (sea.mines < 1 || sea.mines > sea.xs * sea.ys - 10) {
      printf("Poet min je mimo rozsah: 1 < (n = %d) < (x * y = %u)\n",
        sea.mines, sea.xs * sea.ys - 10);
      sea_free(&sea);
      return 7;
    }
    sea_randomize(sea);
  }
  sea_count_mines(sea);
  sea_hide(sea);
 
  char line_buf[81];
  char *string, a;
  int x, y, stat = 0;

  for (;;) {
    switch (stat) {
      case 2: sea_flag_mines(sea); break;
      case 1: sea_uncover_mines(sea); break;
      default: break;
    }
    sea_print(sea);
    switch (stat) {
      case 2:
        printf("HUR! MOE JE PROZKOUMNO!\n");
        break;
      case 1:
        printf("BUM!!\n");
        break;
      case 0:
        printf("Napite ?, vype se nvod.");
        break;
      case -1:
        printf("[%d,%d] je ji odkryto.", x, y);
        break;
      case -2:
        printf("[%d,%d] je ji oznaeno.", x, y);
        break;
      case -3:
        printf("[%d,%d] je mimo moe.", x, y);
        break;
      case -4:
        printf("patn pkaz. Zkuste napsat otaznk, vype se npovda.");
        break;
    }

    if (stat > 0) break;
    
    printf(" | >>> ");
    if ((string = strip(fgets(line_buf, sizeof(line_buf), stdin))) == NULL)
      break;

    stat = 0;
    
    if (!strcmp(string, "?")) {
      printf("<sloupec> <dek>   -- odkryje zadan polko\n"
             "<sloupec> <dek> f -- ozna zadan polko\n"
             "<sloupec> <dek> x -- ozna polko i bezprostedn okol\n"
             "exit                -- ukon program\n"
             "?                   -- vype tuto npovdu\n"
             "\n"
             "Souadnice polek jsou slovny od jedniky!\n"
             "\n"
             "Stisknte ENTER.\n");
      while (fgetc(stdin) != '\n');
    } else
      if (!strcmp(string, "exit")) break;
      else {
        switch (sscanf(string, "%d %d %c", &x, &y, &a)) {
          case 2: // jen odkryt
            stat = sea_uncover(sea, x-1, y-1);
            break;
          case 3: // odkryt/znaka
            switch (a) {
              case 'x':
                stat = sea_uncover_block(sea, x-1, y-1);
                break;
              case 'f':
                stat = sea_flag(sea, x-1, y-1);
                break;
              default:
                stat = -4;
                break;
            }
            break;
          default:
            stat = -4;
            break;
        }
      }
  }
  printf("Konec...\n");
  sea_free(&sea);
  return 0;
}
