#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <locale.h>

/*
 *  P R  C E   S E   S E Z N A M E M
 */

/* struktura pro seznam i s potadlem dk */
typedef struct {
  char **lines;
  int n;
} LIST;

/* nastav nulov hodnoty */
void newlist(LIST *list) {
  list->lines = NULL;
  list->n = 0;
}

/* pid dal prvek do seznamu, kter nejprve nathne */
int addline(LIST *list, char *data) {
  int length;
  char **lines2;
  /* nathne seznam */
  if ((lines2 = (char **) realloc(list->lines,
    (list->n + 1) * sizeof(char*))) == NULL)
    return -1;
  list->lines = lines2;
  /* msto nulovch etzc bude lep mt etzce przdn */    
  if (data == NULL)
    length = 0;
  else
    length = strlen(data);
  /* kontrola alokace msta pro etzec */
  if ((list->lines[list->n] = (char *) malloc(length + 1)) == NULL)
    return -2;
  if (data != NULL)
    strcpy(list->lines[list->n], data);
  else
    list->lines[list->n][0] = '\0';
  /* piteme jeden dek */    
  list->n++;    
  return 0;
}

/* odstran seznam z pamti */
void freelist(LIST *list) {
  int i;
  if (list == NULL)
    return;
  for (i = 0; i < list->n; i++) {
    if (list->lines[i] != NULL)
      free(list->lines[i]);
  }
  free(list->lines);
  /* klid */
  newlist(list);
}

/*
 *  Q U I C K S O R T
 */

/* Podle pramen na sti m sloitost O(n^2) v nejnepznivjch ppadech,
 * vtinou to ale je O(n . log n) . */
 
/* Setd prvky od prvnho po posledn s vyuitm knihovn funkce strcoll 
 * k porovnn etzc s ohledem na nrodn abecedy. */
void quicksort(char **lines, int first, int last)
{
  char *tmp, *pivot;
  /* potebn odkazy */
  int up, down;

  /* zdrovat se nebudeme */
  if (first >= last) return;

  /* pivot je dlic prvek, nejprve je nastaven na libovolnou hodnotu */
  pivot = lines[(first+last) / 2];
  /* ukazatele na nejbli prvky vt, i men ne dlic prvek */
  up = first;
  down = last;
  do {
    /* kter je nejbli vt ne dlic prvek? */
    while (strcoll(lines[up], pivot) < 0) up++;
    /* a kter je nejbli men? */
    while (strcoll(lines[down], pivot) > 0) down--;
    /* vt je ped menm, prohodme je a posuneme se */
    if (up <= down)
    {
      /* ppad, kdy bude prohazovn zbyten, by mohlo bt
       * mn, ne tch, kdy by bylo zbyten to zjiovat */     
      tmp = lines[up];
      lines[up] = lines[down];
      lines[down] = tmp;
      up++;
      down--;
    }
  } while (up <= down); /* je to poad pod patn? */
  /* doln prvek by ml bt prvn, pokud nen, jet je co tdit */
  if (first < down) quicksort(lines, first, down);
  /* a horn by ml bt zase nahoe */
  if (up < last) quicksort(lines, up, last);
}

/*
 *  H L A V N    P R O G R A M
 */

#define LINE_MAX 1024

#define ERROR_SYNTAX        1
#define ERROR_NO_MEMORY     2
#define ERROR_LINE_TOO_LONG 3
#define ERROR_IO_ERROR      4

int main(int argc, char **argv) {
  FILE *f;
  int noclose = 0;
  char line_buf[LINE_MAX+2], *p;
  int state = 0;
  int i;

  LIST list;

  setlocale(LC_ALL, "");
  
  if (argc < 2) {
    printf("nvod: sort textov_soubor_k_setdn\n"
	   "Chcete-li tdit data ze standardnho vstupu, zadejte pomlku.\n"
	   "\n"
	   "Nvratov hodnoty:\n"
	   "\t1 - patn parametry, vypsala se npovda\n"
	   "\t2 - dola pam\n"
	   "\t3 - pli dlouh dek (max. %d znak)\n"
	   "\t4 - chyba pi natn souboru\n",
	   LINE_MAX);
    return ERROR_SYNTAX;
  }

  if (!strcmp(argv[1], "-")) {
    f = stdin; /* pomlka zna standardn vstup */
    noclose = 1;
  }
  else
    f = fopen(argv[1], "r"); /* tohle by ml bt obyejn soubor */

  newlist(&list);

  /* nati vechny dky */
  while (!feof(f)) {
    /* toto je dost zsadn krok, fgets etzec na konci souboru nevynuluje
       a my nechceme mt posledn dek na vstupu dvakrt :) */
    line_buf[0] = '\0';
    /* jeden dek... */
    if (fgets(line_buf, sizeof(line_buf), f) == NULL) {
      if (!feof(f)) { /* toto je u chyba */
        state = ERROR_IO_ERROR;
        perror("Chyba");
        break;
      }
    }
    if (line_buf[0] == '\0') continue; /* tohle smet nechceme... */
    
    /* Kad dek m konit na '\n', jen na konci souboru se to
     * stt nemus. Spravme to. */
    if ((p = strrchr(line_buf, '\0')) != NULL && --p >= line_buf
      && *p != '\n') strcat(line_buf, "\n");
    
    if (strlen(line_buf) > LINE_MAX) {
      fprintf(stderr,
        "%d. dek je moc dlouh (max. %d znak). Rovn nen doporueno "
        "tmto\n"
        "programem zpracovvat binrn soubory.\n",
        list.n+1, LINE_MAX);
      state = ERROR_LINE_TOO_LONG;
      break;	      
    }
    
    /* ulo do seznamu */
    if (addline(&list, line_buf) < 0) { /* nepovedlo se? */    
      fprintf(stderr, "Zd se, e se soubor do pamti nevejde...\n");
      freelist(&list); /* aspo uvolnme to, co jsme tam dostali */
      state = ERROR_NO_MEMORY;
      break;
    }
  }
  
  /* soubor zaveme, pokud nen stdin */
  if (!noclose)
    fclose(f);
  
  if (state == 0) {
    /* setdme */
    quicksort(list.lines, 0, list.n - 1);
  
    if (list.lines != NULL) /* ist pro jistotu, kdyby byl soubor przdn */
      for (i = 0; i < list.n; i++) {
        printf("%s", list.lines[i]);
      }
  }

  /* uklidme */
  freelist(&list);
  
  /* pochlubme se vsledkem */
  return state;
}
