#include <stdio.h>
#include <string.h>

#include "bitmap.h"
#include "bitmap_write.h"
#include "bitmap_libart.h"

#define E_OK                0x00
#define E_SYNTAX            0x01
#define E_CREATE_FAILED     0x02
#define E_BAD_WIDTH         0x03
#define E_BAD_HEIGHT        0x04
#define E_BITMAP_FAILED     0x05
#define E_INVALID_EXTENSION 0x06
#define E_READ_FAILED       0x10
#define E_WRITE_FAILED      0x20
#define E_BROKEN_DATA       0x40
#define E_INVALID_DATA      0x80

int main(int argc, char **argv) {
  FILE *f;
  BITMAP *bmp;
  int width, height;
  char *p;
  int status = E_OK;
  char line[1024];
  char *img_type;
  int img_type_id = -1;

  if (argc < 4) {
    printf("Nvod: circle ka_obrzku vka_obrzku obrzek.ppona\n");
    printf("Pponou je uren formt vstupnho souboru, pouijte png i jpg.\n");
    return E_SYNTAX;
  }
  
  /* vyrovnvac pam nechceme, protoe je dobr vidt i nedopsan dky */
  setbuf(stdout, NULL);
  
  /* ka */
  width = strtol(argv[1], &p, 0);
  if (*p != '\0' || width <= 0) {
    printf("Neplatn ka '%s'\n", argv[1]);
    return E_BAD_WIDTH;
  }
  /* vka */
  height = strtol(argv[2], &p, 0);
  if (*p != '\0' || height <= 0) {
    printf("Neplatn vka '%s'\n", argv[2]);
    return E_BAD_HEIGHT;
  }

  /* Urme typ souboru. Nejprve si vimneme lomtek v nzvu. */
  p = strrchr(argv[3], '/');
  if (p == NULL)
    p = argv[3];
  else
    p++;
  /* Te najdeme pponu. */
  img_type = strrchr(p, '.');
  if (img_type == NULL ) {
    printf("Jmno souboru '%s' neobsahuje pponu!\n", p);
    return E_INVALID_EXTENSION;
  }
  img_type++;
  if ((img_type_id = bitmap_write_match_type(img_type)) < 0) {
    printf("Ppona '%s' je neplatn.\n", img_type);
    return E_INVALID_EXTENSION;
  }
  
  printf("Vytvm vstupn soubor '%s': ", argv[3]);
  if ((f = fopen(argv[3], "wb")) != NULL) {
    printf("vytvoen.\n");
    printf("Vytvm prostor pro obrzek %dx%d: ", width, height);
    if ((bmp = createbitmap2(width, height, 24, 4)) != NULL) {
      int x, y, r;
      printf("hotovo.\n");
      printf("Kreslm...\n");
      while (!feof(stdin)) {
        line[0] = '\0'; /* itn dku */
        if (fgets(line, sizeof(line), stdin) == NULL) { /* kontrola chyb */
	  if (!feof(stdin)) { /* ped koncem souboru je to opravu chyba */
	    perror("Chyba ten");
	    status |= E_READ_FAILED;
	    break;
	  }
	}
	if (line[0] == '\0') continue; /* peskome przdn dky */
        if (sscanf(line, "%d %d %d", &x, &y, &r) != 3) {
	  status |= E_BROKEN_DATA;
	  printf("dek '%s' je neplatn a bude peskoen.\n", line);
	} else { /* krunice bude bl s ernm vnitkem */
	  if (x < 0 || y < 0 || r < 0) { /* zporn hodnoty ne */
	    printf("Jedna z hodnot (x,y)=(%d,%d), r = %d je zporn.\n",
	      x, y, r);
	    status |= E_INVALID_DATA;
          } else {
  	    printf("Vykresluji (x,y)=(%d,%d), r = %d: ", x, y, r);
	    if (bitmap_circle_artsvp(bmp, x, y, r, 0, 0xffffff) == 0)
	      printf("hotovo.\n");
	    else {
	      printf("nesprvn hodnoty, obrzek me bt pokozen.\n");
	      status |= E_INVALID_DATA;
	    }
	  }
	}
      }
      printf("Ukldm obrzek: ");
      if (bitmap_write(bmp, f, img_type_id) == 0)
        printf("v podku.\n");
      else {
        printf("nezdailo se.\n");
        status |= E_WRITE_FAILED;
      }
      /* uvolnn pamti */
      freebitmap(bmp);
    } else {
      printf("nezdailo se.\n");
    }
    fclose(f);
  } else {
    perror("chyba");
    return E_CREATE_FAILED;
  }

  if (status == 0)
    printf("Vykreslovn je dokoneno.\n");
  else
    printf("Vykreslovn skonilo, ale s chybami.\n");
  return status;
}
