// C99

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <netdb.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <arpa/inet.h>

// hack pro aisu
#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0
#endif

#define UNITS_MAX 3
static struct { char *format; int div; } units[UNITS_MAX] = {
  { "%.1f MB", 1048576 }, { "%.1f kB", 1024 }, { "%.0f B", 1 } };
char *calculate_size(double count, char *out, int out_limit) {
  int i;
  for (i = 0; i < UNITS_MAX - 1 && count < units[i].div; i++);
  snprintf(out, out_limit, units[i].format, count / units[i].div);
  return out;
}

#define BUFFER_SIZE     262144
// max. pouiteln hodnota pro TCP
#define TCP_BUFFER_SIZE 262144 
// max. pouiteln hodnota pro UDP
#define UDP_BUFFER_SIZE 32768 
// poten hodnota pro klienta
#define CHUNK_SIZE      1024
// doln mez
#define THRESHOLD_SIZE  512

// Serverov modul

// Stavov hodnoty
#define S_NOSOCK 1
#define S_NOSOPT 2
#define S_NOBIND 3
#define S_NOLSTN 4
#define S_NOACPT 5

char *s_errors[] = {
  "bez chyby", "nelze vytvoit zsuvku", "nelze nastavit parametry zsuvky",
  "nelze nastavit adresu zsuvky", "nelze nastavit pjem",
  "nelze pijmat spojen" };
  
// port    - slo portu
// in_addr - ukazatel na strukturu s adresou, na kter m server poslouchat
// udp     - 1 pro UDP, 0 pro TCP spojen
int servermod(int port, struct in_addr const *in_addr, int udp) {
  struct sockaddr_in sockaddr, udpsockaddr;
  socklen_t sockaddr_len;
  int sock, fd = 0;
  long int i;
  char ipv4as[INET_ADDRSTRLEN];
  char tmp2[32];
  char buffer[BUFFER_SIZE];

  int nr = 0;
  long int udpcount = 0;
  double count1, count2, timer1, timer2;
  struct timeb t1, t2;
 
  /* nov zsuvka */
  if ((sock = udp ? socket(PF_INET, SOCK_DGRAM , IPPROTO_UDP) :
                    socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
    return S_NOSOCK;
  /* chceme adresu hned te */
  i = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) < 0)
    return S_NOSOPT;

  memset(&sockaddr, 0, sizeof(struct sockaddr_in)); // oh shit!
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_port = htons(port);
  sockaddr.sin_addr = *in_addr;

  /* nastavme adresu pro pjem */
  if (bind(sock, (struct sockaddr*)&sockaddr, sizeof(struct sockaddr_in)) < 0)
    return S_NOBIND;
  
  inet_ntop(AF_INET, in_addr, ipv4as, sizeof(ipv4as));
  printf("Adresa serveru je %s:%d.\n", ipv4as, port);
  
  for (;;) {
    if (!udp) { // TCP reim
    if (listen(sock, 1) < 0)
        return S_NOLSTN;
      sockaddr_len = sizeof(struct sockaddr_in);
      if ((fd = accept(sock, (struct sockaddr*)&sockaddr, &sockaddr_len)) < 0) {
        printf("Pi pipojovn klienta dolo k chyb.\n");
        return S_NOACPT;
      }
      inet_ntop(AF_INET, &sockaddr.sin_addr, ipv4as, sizeof(ipv4as));
      printf("TCP spojen od %s:%d\n", ipv4as, ntohs(sockaddr.sin_port));
    }

    count1 = count2 = 0;
    ftime(&t1);
    timer2 = 0;
    for (;;) {
      if (!udp) // TCP
        nr = recv(fd, buffer, sizeof(buffer), 0);
      else { // UDP
        sockaddr_len = sizeof(struct sockaddr_in);
        nr = recvfrom(sock, buffer, sizeof(buffer), 0,
          (struct sockaddr *) &sockaddr, &sockaddr_len);
        if (memcmp(&udpsockaddr, &sockaddr, sizeof(struct sockaddr_in))) {
          inet_ntop(AF_INET, &sockaddr.sin_addr, ipv4as, INET_ADDRSTRLEN);
          printf("\nUDP spojen od %s:%d\n", ipv4as, ntohs(sockaddr.sin_port));
          udpsockaddr = sockaddr;
          udpcount = 0;
        }
        if (nr > 0) // na zatku bylo slo paketu
          sendto(sock, buffer, sizeof(long int), MSG_NOSIGNAL,
            (struct sockaddr *) &sockaddr, sockaddr_len);
        udpcount++;
      }
      ftime(&t2);
      timer1 = (double) (t2.time - t1.time) +
        (double) (t2.millitm - t1.millitm) / 1000;
      if (nr == 0) { // podle man. strnky to znamen, e se spojen ukonilo
        printf("\nSpojen bylo ukoneno.\n");
        break;
      }
      if (nr < 0) { // dolo k chyb - konec
        printf("\nSpojen bylo perueno (kd #%d: %s)\n",
	  errno, strerror(errno));
        break;
      }
      count1 += nr;
      count2 += nr;
      printf("Pijato %s, ", calculate_size(nr, tmp2, sizeof(tmp2)));
      printf("celkem %s za %.1f s => ", 
        calculate_size(count1, tmp2, sizeof(tmp2)), timer1);
      printf("rychlost: %s/s prmrn",
        calculate_size((double) count1 / timer1, tmp2, sizeof(tmp2)));
      if (timer1 - timer2 > 1)
        printf(", %s/s okamit\n",
          calculate_size((double) count2/(timer1-timer2), tmp2, sizeof(tmp2)));
      else
        printf("    \r");
      if (timer1 - timer2 > 1) {
        timer2 = timer1;
        count2 = 0;
      }
    }
    printf("\n");
  }
  
  close(fd);
  close(sock);
  return 0;
}

// Klientsk modul

// stavov hodnoty
#define C_NOSOCK 1
#define C_NOCONN 2

// port    - slo portu
// in_addr - ukazatel na strukturu s adresou, na kter m server poslouchat
// udp     - 1 pro UDP, 0 pro TCP spojen
// rnd     - odesln nhodnch dat
int clientmod(int port, struct in_addr const *in_addr, int udp, int rnd) {
  struct sockaddr_in sockaddr;
  socklen_t sockaddr_len;
  int sock;
  long int i;
  char ipv4as[INET_ADDRSTRLEN];
  char tmp2[32];
  char buffer[BUFFER_SIZE];
  const int buffer_size = udp ? UDP_BUFFER_SIZE : TCP_BUFFER_SIZE;
  int chunk_size = CHUNK_SIZE;
  long int udpcount = 0;
  fd_set fds1, fds2;

  int nr = 0;
  struct timeb t1, t2;
  double count1, count2, timer1, timer2;
 
  /* nov zsuvka */
  if ((sock = udp ? socket(PF_INET, SOCK_DGRAM , IPPROTO_UDP) :
                    socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
    return C_NOSOCK;
  
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_port = htons(port);
  sockaddr.sin_addr = *in_addr;

  inet_ntop(AF_INET, in_addr, ipv4as, sizeof(ipv4as));
  printf("Adresa serveru je %s:%d.\n", ipv4as, port);
  
  if (!udp) { // TCP reim
    if (connect(sock, (struct sockaddr *) &sockaddr,
      sizeof(struct sockaddr_in)) < 0) {
      printf("Pi pipojovn klienta dolo k chyb.\n");
      return C_NOCONN;
    }
  }
 
  FD_ZERO(&fds1);
  FD_SET(sock, &fds1);
  
  count1 = count2 = 0;
  timer2 = 0;
  ftime(&t1);
  for (;;) {
    if (rnd) // tvorba odpadu
      for (i = 0; i < chunk_size; i++) buffer[i] = rand();
    if (!udp) // TCP
      nr = send(sock, buffer, chunk_size, MSG_NOSIGNAL);
    else { // UDP
      i = -1;
      while (i != udpcount) {
        // poleme data, na zatek hodme znaku
        * (long int *) buffer = udpcount;
        nr = sendto(sock, buffer, chunk_size, MSG_NOSIGNAL,
          (struct sockaddr *) &sockaddr, sizeof(struct sockaddr_in));
        fds2 = fds1;
        select(sock+1, &fds2, NULL, NULL, &(struct timeval) { 0, 333333 }); // 1 s
        if (FD_ISSET(sock, &fds2)) // nco pilo
          // zkontrolujeme
          recvfrom(sock, &i, sizeof(long int), 0, (struct sockaddr *) &sockaddr,
            &sockaddr_len);
      }
      udpcount++;
    }
    if (nr == 0) { // podle man. strnky to znamen, e se spojen ukonilo
      printf("\nSpojen bylo ukoneno.\n");
      break;
    }
    if (nr < 0) { // dolo k chyb - konec
      printf("\nSpojen bylo perueno (kd #%d: %s)\n", errno,strerror(errno));
      break;
    }
    ftime(&t2);
    timer1 = (double) (t2.time - t1.time) +
      (double) (t2.millitm - t1.millitm) / 1000;
    count1 += nr;
    count2 += nr;
    printf("Odeslno %s, ", calculate_size(nr, tmp2, sizeof(tmp2)));
    printf("celkem %s za %.1f s => ",
      calculate_size(count1, tmp2, sizeof(tmp2)), timer1);
    printf("rychlost: %s/s prmrn",
      calculate_size((double) count1 / timer1, tmp2, sizeof(tmp2)));
    if (timer1 - timer2 > 1)
      printf(", %s/s okamit\n",
        calculate_size((double) count2 / (timer1-timer2), tmp2, sizeof(tmp2)));
    else
      printf("    \r");
    // nasteven bloku pro odesln
    if (timer1 > 2) { // neunhlujme se
      chunk_size = count1 / (2*timer1);
      if (chunk_size > buffer_size) chunk_size = buffer_size;
      if (chunk_size < THRESHOLD_SIZE) chunk_size = THRESHOLD_SIZE;
    }
    if (timer1 - timer2 > 1) {
      timer2 = timer1;
      count2 = 0;
    }
  }
  printf("\n");
  
  close(sock);
  return 0;
}

#define BAD_ARGS 1
int main(int argc, char **argv) {
  int port = 0;
  struct in_addr in_addr;
  int i;
  int argc2 = 0; /* potadlo parametr bez pomlek */
  char *p;
  int protocol = IPPROTO_TCP;
  int client = 0, rnd = 0;
  
  if (argc < 2) {
    printf("nvod: bandwidth port [adresa] [--tcp|--udp] [--rnd]\n"
      "Je-li zadno jen slo portu, program bude ekat na spojen.\n"
      "Pokud je navc zadna adresa, pokus se na ni program pipojit.\n"
      "Jet je mon zvolit protokol, kter bude pouit. TCP je vchoz.\n"
      "Pot, co je spojen navzno, odesl klient sam nuly serveru. To\n"
      "me nap. pi pouit spoje s kompres zkreslit men. Parametrem\n"
      "--rnd tomu lze zabrnit, program bude odeslat nhodn znaky, kter\n"
      "se komprimovat nedaj.\n");
    return BAD_ARGS;
  }

  setbuf(stdout, NULL); // chceme vypisovat i neukonen dky 
 
  srand(time(0)); 
  
  /* vchoz adresa */
  in_addr.s_addr = INADDR_ANY;
  
  for (i = 1; i < argc; i++) {
    if (argv[i][0] != '-') { // bez pomlky 
      switch (argc2) {
        case 0:
          p = NULL;
          port = strtol(argv[i], &p, 10);
          if (argv[i][0] == '\0' || *p != '\0') {
            printf("Neplatn slo portu: '%s'\n", argv[i]);
            return BAD_ARGS;
          }
          break;
        case 1: {
            struct hostent *he;
            if ((he = gethostbyname(argv[i])) != NULL)
              memcpy(&in_addr, he->h_addr_list[0], sizeof(struct in_addr));
            else {
              printf("Neplatn adresa: '%s'\n", argv[i]);
              return BAD_ARGS;
            }
            client = 1;
	  }
          break;
        default:
          break;
      }
      argc2++;
    } else {
      if (!strcmp(argv[i], "--tcp")) // TCP
        protocol = IPPROTO_TCP;
      if (!strcmp(argv[i], "--udp")) // UDP
        protocol = IPPROTO_UDP;
      if (!strcmp(argv[i], "--random")) // odeslat odpad msto nul
        rnd = 1;
    }
  }
  printf("Protokol: %s\n", protocol == IPPROTO_TCP ? "TCP" : "UDP");
  
  if (client) {
    printf("Program pracuje v klientskm reimu a bude odeslat %sdata.\n",
      rnd ? "nhodn " : "");
    if (protocol == IPPROTO_UDP)
      printf("Men rychlosti UDP serveru na stran klienta nen "
        "spolehliv.\nSledujte men na serveru.\n");
    i = clientmod(port, &in_addr, protocol == IPPROTO_UDP, rnd);
  } else {
    printf("Program pracuje v servervm reimu a oekv data.\n");
    i = servermod(port, &in_addr, protocol == IPPROTO_UDP);
  }
  if (i != 0) {
    printf("Chyba: %s\n", s_errors[i]);
    printf("Podle systmu: %s (%s)\n", strerror(errno),
      errno == EBADF ? "EBADF" :
      errno == ENOTSOCK ? "ENOTSOCK" :
      errno == EADDRNOTAVAIL ? "EADDRNOTAVAIL" :
      errno == EADDRINUSE ? "EADDRINUSE" :
      errno == EINVAL ? "EINVAL" :
      errno == EACCES ? "EACCES" :
      errno == EFAULT ? "EFAULT" :
      errno == EACCES ? "EACCES" : "jin chyba");
  }  
  return 0;
}
