/* NVRH ZADN LOHY:
 * 
 * Pi prci v unixu lze asto narazit na potebu prohldnout si njak text.
 * Me jt jak o obyejn textov soubor, tak o vstup njakho programu.
 * K tomuto elu se dlouho pouval program 'more', kter po stisku klvesy
 * Enter postupn text ze zadanho souboru, nebo standardnho vstupu.
 * 
 * Za kol je napsat program, kter bude tuto funkci plnit. K uloen textu
 * v pamti je doporueno pout dynamicky alokovanou pam pomoc funkc
 * malloc a realloc. Protoe nen mon zjistit dlku vstupnch dat pi
 * natn ze standardnho vstupu, bude nutn pro zvtovn alokovanho
 * kusu pamti pout realloc.
 *
 * Pi natn textu by bylo ikovn, kdyby nebyla dlka dku nikterak
 * omezen. To lze zadit na pklad natnm souboru po jednotlivch
 * znacch vdy a po konec dku, tento zpsob je vak pomrn tkopdn.
 * Bylo by vhodn pout efektivnj zpsob. (+1 bod)
 * 
 * Jako voliteln, nicmn doporuen rozen je monost pevjen textu
 * vped a vzad. V takovm ppad bude vhodn pout knihovnu ncurses. Bude
 * se hodit zejmna ke zjiovn kd stisknutch klves, protoe funkce
 * standardn knihovny jako fgetc ekaj na konec dku, tj. stisk klvesy
 * Enter. To by siln znepjemovalo ovldn programu. (+1 bod za pevjen,
 * +1 bod za pouit curses)
 *
 * Dalm monm rozenm je rolovn do stran. Zalomen dky toti
 * nepsob pkn. (+1 bod)
 *
 * Jako obvykle lze zskat jet dva body navc za odevzdn do konce tdne po
 * zadvacm cvien.
 *
 * Poznmky:
 *
 * Pro sprvnou funkci je poteba znt rozmry terminlu. Pokud program
 * nebude pouvat curses, ml by si je zjistit sm z promnnch prosted
 * COLUMNS a LINES. Pi jejich neptomnosti lze pedpokldat obvykl rozmry
 * 80x24.
 *
 * Pokud program nebude podporovat rolovn do stran, ml by dky zalamovat.
 * Pak je ale teba dt pozor na to, e jeden dek v souboru me zabrat
 * nkolik dk na terminlu, nebo e se nemus ani vejt na strnku.
 *
 * K een lohy je vhodn vyut knihovnu curses, je vak mono zvolit
 * kteroukoli jinou poskytujc potebn sluby.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <malloc.h>
#include <stdlib.h>
#include <curses.h>
#define BLOCK 16384 
int main(int argc, char **argv) {
  FILE *input = argc > 1 ? fopen(argv[1], "rb") : stdin;
  if (input == NULL) {
    perror("Chyba na vstupu");
    return 1;
  }
  char *data = NULL, **lp;
  int c = 0;
  long int size = 0, size2, lc = 1, i, j, k = 0, l = 0;
  WINDOW *w1, *w2;
  while (!feof(input)) {
    if ((data = (char *) realloc(data, size + BLOCK)) == NULL) {
      fprintf(stderr, "Out of memory (%ld B)\n", size + j);
      if (input != stdin) fclose(input);
      return 2;
    }
    size2 = size;
    j = BLOCK;
    while ((i = fread(data+size, 1, j, input)) > 0 && (size += i, i<j)) j -= i;
    for (i = size2; i < size; i++) if (data[i] == '\n') lc++;
  }
  if (input != stdin) fclose(input);
  if ((lp = (char **) malloc(lc * sizeof(char *))) == NULL) {
    free(data);
    fprintf(stderr, "Out of memory (lookup array, %ld B)\n", lc);
    return 3;
  }
  if (data[size-1] == '\n') lc--;
  lp[j = 0] = data;
  for (i=0;i<size;i++) if (data[i]=='\n' && i < size) lp[++j] = data+i+1;
  close(0);
  open("/dev/tty", O_RDONLY); // pkn nzkorovov pesmrovn
  initscr(); cbreak(); noecho(); // pprava knihovny curses
  w1 = newwin(LINES - 1, COLS, 0, 0);
  scrollok(w1, FALSE);
  w2 = newwin(1, COLS, LINES - 1, 0);
  scrollok(w2, FALSE);
  keypad(w2, TRUE);
  wprintw(w2, "Q = quit, H = left, J = down, K = up, L = right");
  wrefresh(w2);
  do {
    switch (c) {
      case 'K': case KEY_UP   : if (k > 0) k--; break;
      case KEY_ENTER:
      case 'J': case KEY_DOWN : if (k + LINES - 1 < lc) k++; break;
      case 'H': case KEY_LEFT : if (l > 0) l--; break;
      case 'L': case KEY_RIGHT: l++; break;
      case KEY_PPAGE: if (k - LINES + 1 > 0) k -= LINES-1; else k = 0; break;
      case KEY_NPAGE:
        if (lc > LINES-1)
          k = k+2*LINES < lc+1 ? k+LINES-1 : lc - LINES + 1; break;
      default: break;
    }
    wclear(w1);
    wmove(w1, 0, 0);
    for (i = 0; i < LINES-1 && k+i < lc; i++) {
      for (j = 0; j < COLS && &lp[k+i][j+l]-data < size && 
        lp[k+i][j+l] != '\n'; j++)
        if (k+i+1 == lc || &lp[k+i][l] < lp[k+i+1])
          wprintw(w1, "%c", lp[k+i][j+l]);
      if (k+i+1 < lc)  
        waddch(w1, '\n');
    }
    wrefresh(w1);
  } while ((c = toupper(wgetch(w2))) != 'Q');
  delwin(w2);
  delwin(w1);
  endwin();
  free(lp);
  free(data);
  return 0;
}
