// chat.c 

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <locale.h>

#ifdef IRIX_MODE
#include <signal.h>
#endif

#include "chat.h"
#include "protocol.h"

char *error_messages[] = { "bez chyby", "nelze vytvoit zsuvku",
  "nelze nastavit parametry zsuvky", "zsuvce nelze nastavit adresu",
  "nelze nastavit pijem na zsuvce", "mlo pamti", "nelze se pipojit",
  "chyba pi pihlen" };

struct setup setup = { .mode = SERVER, .nick = NULL };

void odpad() {
  if (setup.nick != NULL) free(setup.nick);
}

int main(int argc, char **argv) {
  char *p;
  atexit(odpad); // pro sprvn ukonen

#ifdef IRIX_MODE
  signal(SIGPIPE, SIG_IGN); // pro jistotu
#endif

  setlocale(LC_ALL, "");
  
  if (argc < 2) {
    printf("nvod: chat slo_portu [adresa] [jmno]\n");
    printf("Je-li zadno jen slo portu, pracuje program jako server.\n");
    printf("Je-li zadna tak clov adresa a jmno, jede jako klient.\n");
    return NOARGS;
  }

  setup.port = strtol(argv[1], &p, 0);
  if (argv[1][0] == '\0' || *p != '\0') { // port
    printf("patn slo portu - '%s'.\n", argv[1]);
    return BADARGS;
  }
  if (argc >= 4) { // adresa
    struct hostent *he = gethostbyname(argv[2]);
    if (he == NULL || he->h_addrtype != AF_INET) {
      printf("Neplatn adresa - '%s'.\n", argv[2]);
      return BADARGS;
    }
    memcpy(&setup.addr, he->h_addr_list[0], sizeof(struct in_addr));
  }
  if (argc >= 4) { // jmno
    if ((setup.nick = (char *) malloc(strlen(argv[3])+1)) == NULL) {
      printf("Nedostatek pamti pro pezdvku.\n");
      return NOMEM;
    }
    strcpy(setup.nick, argv[3]);
    setup.mode = CLIENT;
  }

  setbuf(stdout, NULL);
  
  int status, i;
  WINDOW *wmessages, *winputbox, *wm2, *wi2;

  switch (setup.mode) {
    case CLIENT:
      initscr(); start_color(); cbreak(); noecho(); // pprava curses
 
      for (i = 0; i < COLOR_PAIRS; i++)
        /*        number        foreground   background */
        init_pair(i,            i % 8,       i / 8);
      
      wmessages = newwin(LINES - 3, COLS, 0, 0);
      winputbox = newwin(3, COLS, LINES - 3, 0);
  
      if (wmessages == NULL || winputbox == NULL) {
        if (wmessages != NULL) delwin(wmessages);
        if (winputbox != NULL) delwin(winputbox);
        return CURSESFAILED;
      }

      wattrset(wmessages, COLOR_PAIR(COLOR_BLUE));
      wborder(wmessages, ACS_VLINE, ACS_VLINE, ACS_HLINE, ACS_HLINE,
        ACS_ULCORNER, ACS_URCORNER, ACS_LLCORNER, ACS_LRCORNER);
      wmove(wmessages, 0, 4);
      waddch(wmessages, ACS_RTEE);
      wattrset(wmessages, COLOR_PAIR(COLOR_YELLOW | A_BOLD));
      wprintw(wmessages, " Pokec ");
      wattrset(wmessages, COLOR_PAIR(COLOR_BLUE));
      waddch(wmessages, ACS_LTEE);

      wmove(wmessages, LINES - 4, COLS - 25);
      waddch(wmessages, ACS_RTEE);
      wattrset(wmessages, COLOR_PAIR(COLOR_BLUE | A_BOLD));
      wprintw(wmessages, " F10 - konec ");
      wattrset(wmessages, COLOR_PAIR(COLOR_BLUE));
      waddch(wmessages, ACS_LTEE);

      wattrset(winputbox, COLOR_PAIR(COLOR_GREEN));
      wborder(winputbox, ACS_VLINE, ACS_VLINE, ACS_HLINE, ACS_HLINE,
        ACS_ULCORNER, ACS_URCORNER, ACS_LLCORNER, ACS_LRCORNER);
      wmove(winputbox, 0, 4);
      waddch(winputbox, ACS_RTEE);
      wattrset(winputbox, COLOR_PAIR(COLOR_YELLOW | A_BOLD));
      wprintw(winputbox, " Text k odesln ");
      wattrset(winputbox, COLOR_PAIR(COLOR_GREEN));
      waddch(winputbox, ACS_LTEE);
  
      wmove(winputbox, 2, COLS - 25);
      waddch(winputbox, ACS_RTEE);
      wattrset(winputbox, COLOR_PAIR(COLOR_GREEN | A_BOLD));
      wprintw(winputbox, " F5/ENTER - poslat ");
      wattrset(winputbox, COLOR_PAIR(COLOR_GREEN));
      waddch(winputbox, ACS_LTEE);
      
      wrefresh(wmessages);
      wrefresh(winputbox);

      wm2 = derwin(wmessages, LINES - 5, COLS - 2, 1, 1);
      wi2 = derwin(winputbox, 1, COLS - 2, 1, 1);
      keypad(wi2, TRUE);
      scrollok(wm2, TRUE);
      scrollok(wi2, FALSE);
      
      status = chat_client(&setup, wm2, wi2);

      delwin(wi2);
      delwin(wm2);
      delwin(wmessages);
      delwin(winputbox);
      endwin();
      if (status != ERR_OK) {
        printf("Chyba klienta #%d: %s\n", status, error_messages[status]);      
        printf("Systm: #%d: %s\n", errno, strerror(errno));
      }
      break;
      
    case SERVER:
      setup.addr.s_addr = INADDR_ANY;
      if ((status = chat_server(&setup)) != ERR_OK)
        printf("Chyba serveru #%d: %s\n", status, error_messages[status]);
      break;
  }

 
  return NOERR;
}
