#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <netdb.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include "chat.h"
#include "protocol.h"

// spojov seznam klient
struct client_list {  
  struct in_addr addr;
  unsigned int port;
  int fd;
  unsigned long int last_message_id;
  struct chat_packet inpacket, outpacket;
  char *nick;
};

// pidvn klienta
struct client_list *add_client(struct client_list *list,
  unsigned long int *count, struct in_addr addr, unsigned int port, int fd,
  unsigned long int lastmsgid, char *nick, int nick_l) {

  if (nick_l <= 0) nick_l = strlen(nick);
  char *newnick;
  if ((newnick = (char *) malloc(nick_l+1)) == NULL) return NULL;
  strncpy(newnick, nick, nick_l);
  newnick[nick_l] = '\0';
 
  struct chat_packet newpacket;
  if (createpacket(&newpacket, MAX_PACKET_SIZE) < 0) return NULL;
  if ((list = (struct client_list *) realloc(list,
    ((*count)+1)*sizeof(struct client_list))) == NULL) {
    freepacket(&newpacket);
    free(newnick);
    return NULL;
  }
  
  list[*count].addr = addr;
  list[*count].port = port;
  list[*count].fd = fd;
  list[*count].last_message_id = lastmsgid;
  list[*count].inpacket = newpacket;
  clearpacket(&list[*count].outpacket);
  list[*count].nick = newnick;
  (*count)++;
  return list;
}

// odebrn klienta
struct client_list *del_client(struct client_list *list, int idx,
  unsigned long int *count) {
  if (list == NULL || count == NULL || *count == 0 || idx < 0
    || idx >= *count) return NULL;
  freepacket(&list[idx].inpacket);
  free(list[idx].nick);
  list[idx] = list[*count - 1]; // dme sem posledn (!) prvek a snme poet
  (*count)--; // na poad prvk natst nezle :)
/*  if (*count == 0) {
    free(list);
    return NULL;
  } else */
  return (struct client_list *) realloc(list,
    *count * sizeof(struct client_list));
}

// spojov seznam zprv
struct message_stack {
  unsigned long int length;
  char *message;
};

struct message_stack *add_message(struct message_stack *list,
  unsigned long int *count,
  const char *nick, int nick_l, const char *message, int message_l) {
  char *newmessage;
  int length;
  if (nick_l <= 0) nick_l = strlen(nick);
  if (message_l <= 0) message_l = strlen(message);
  length = nick_l + message_l + 1;
  if ((newmessage = (char *) malloc(length)) == NULL) return NULL;
  if ((list = (struct message_stack *) malloc(
    sizeof(struct message_stack) * (*count + 1))) == NULL) {
    free(newmessage);
    return NULL;
  }
  list[*count].length = length;
  list[*count].message = newmessage;
  memcpy(list[*count].message, nick, nick_l);
  list[*count].message[nick_l] = '\0';
  memcpy(&list[*count].message[nick_l+1], message, message_l);
  printf("New message #%ld (length %d)", *count, message_l);
  if (list[*count].message[0] == 0)
    printf(" - special, code #%d", (int)list[*count].message[1]);
  printf("\n");
  (*count)++;
  return list;
}

// speciln zprvy
const char special_codes[][SPEC_CODE_LEN] = {
  { 0, LOGIN_CODE },  // zprva o pihlen
  { 0, LOGOUT_CODE }, // zprva o odhlen
  };

// spust server 
int chat_server(const struct setup *setup) {
  int s; // vytvoen zsuvky
  struct sockaddr_in sockaddr;
  if ((s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
    return ERR_SOCKET;
  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &(int){1}, sizeof(int)) < 0)
    return ERR_SOCKOPT;
  memset(&sockaddr, 0, sizeof(struct sockaddr_in)); // oh shit!
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_port = htons(setup->port);
  sockaddr.sin_addr = setup->addr;
  if (bind(s, (struct sockaddr *) &sockaddr, sizeof(struct sockaddr_in)) < 0)
    return ERR_BIND;
  if (listen(s, 10) < 0)
    return ERR_LISTEN;

  // seznam klient a jejich zsuvek
  struct client_list *cl = NULL;
  // seznam zprv
  struct message_stack *ml = NULL;
  unsigned long int message_id = 0;
  unsigned long int cn = 0;
  struct pollfd *ufds; // prvn je vdy zsuvka serveru
  int cs; // klientova zsuvka
  if ((ufds = (struct pollfd *) malloc(sizeof(struct pollfd))) == NULL) {
    close(s);
    return ERR_NOMEM;
  }
  ufds[0].fd = s;
  ufds[0].events = POLLIN;

  struct chat_packet inpacket; // balek dat
  if (createpacket(&inpacket, MAX_PACKET_SIZE) < 0) {
    free(ufds);
    close(s);
    return ERR_NOMEM;
  }
  
  printf("Server: %s:%d\n", inet_ntoa(setup->addr), setup->port);
  
  int status, i, j;
  char text_buf[MAX_PACKET_SIZE];
  struct hostent *he;
  
  while (1) {
    printf("Klienti: %p (%lu), zprvy: %p (%lu)\n", (void*)cl, cn,
      (void*)ml, message_id); 
    printf("ek se na data - %lu zsuvek\n", cn+1);
    poll(ufds, cn+1, -1);
    if (ufds[0].revents & POLLIN) { // chce se nkdo pipojit
      printf("Nkdo se pipojuje: ");
      if ((cs = accept(s, (struct sockaddr *) &sockaddr,
        &(socklen_t){sizeof(struct sockaddr_in)})) >= 0) {
        printf("%s:%d, ", inet_ntoa(sockaddr.sin_addr),
          ntohs(sockaddr.sin_port));
        while ((status = recv_packet(cs, &inpacket)) == 0); // pezdvka
        if (status < 0) {
          printf("njak se to pokazilo... #%d: %s\n", errno, strerror(errno));
          close(cs);
        } else { // vechno je v podku, tak uivatele zaadme
          printf(">> ");
          /* slo zprvy, kterou m klient dostat, je o jedniku vy,
           * protoe budeme poslat vem klientm zprvu o pihlen */
          cl = add_client(cl, &cn, sockaddr.sin_addr, sockaddr.sin_port, cs,
            message_id, inpacket.data, inpacket.length);
          clearpacket(&inpacket);
          if (cl != NULL && (ufds = (struct pollfd *) realloc(ufds,
              (cn+1)*sizeof(struct pollfd))) != NULL) {
            printf("[%s]\n", cl[cn-1].nick);
            ufds[cn].fd = cs;
            ufds[cn].events = POLLIN;
            ufds[cn].revents = 0;
            // Poleme nejen jmno uivatele, ale i jeho adresu, nejlpe i
            // reverzn DNS, kdy to pjde.
            he = gethostbyaddr(&cl[cn-1].addr, sizeof(struct in_addr), AF_INET);
            if (he != NULL)
              sprintf(text_buf, "%s na %s (%s)", cl[cn-1].nick,
                inet_ntoa(cl[cn-1].addr), he->h_name);
            else
              sprintf(text_buf, "%s na %s", cl[cn-1].nick,
                inet_ntoa(cl[cn-1].addr));
            ml = add_message(ml, &message_id, special_codes[LOGIN_CODE],
              SPEC_CODE_LEN, text_buf, 0); // pidme zprvu o pihlen
//            continue; // je poteba oberstvit daje o zsuvkch
          } else cl = del_client(cl, cn-1, &cn);
        }
      }
    }

    for (i = 1; i <= cn; i++) {
      printf("Poll #%d: %x\n", i, ufds[i].revents);
      if (ufds[i].revents & POLLIN) { // pila data, nebo dolo k odpojen?
        printf("Zprva od %d. klienta [%s] (%s:%d)\n", i, cl[i-1].nick,
          inet_ntoa(cl[i-1].addr), htons(cl[i-1].port));
        status = recv_packet(cl[i-1].fd, &cl[i-1].inpacket);
        if (status < 0 && errno == EPIPE) { // odpojen
          printf("%d. klient [%s] (%s:%d) se odpojil.\n", i, cl[i-1].nick,
            inet_ntoa(cl[i-1].addr), htons(cl[i-1].port));
          // vymaeme jej ze seznamu
          printf("Poslme zprvu o odhlen\n");
          ml = add_message(ml, &message_id, special_codes[LOGOUT_CODE],
            SPEC_CODE_LEN, cl[i-1].nick, 0); // pidme zprvu o odhlen
          printf("Maeme\n");
          ufds[i] = ufds[cn];
          ufds = (struct pollfd *) realloc(ufds, 
            (cn)*sizeof(struct pollfd));
          cl = del_client(cl, i-1, &cn); // tahle funkce je pkn ujet...
          printf("Vymazno\n");
        }
        if (status > 0) { // u pila cel zprva
          printf("Text zprvy: ");
          for (j = 0; j < cl[i-1].inpacket.length; j++)
            putchar(((char*)cl[i-1].inpacket.data)[j]);
          putchar('\n');
          ml = add_message(ml, &message_id, cl[i-1].nick, 0,
            cl[i-1].inpacket.data, cl[i-1].inpacket.length); // pidme ji
          clearpacket(&cl[i-1].inpacket);
        }
      }
    } 
    for (i = 1; i <= cn; i++) { // Pro to dlit? Teba kvli odhlenm...
      printf("[%s] m posledn zprvu #%ld\n", cl[i-1].nick,
        cl[i-1].last_message_id);
      if (message_id > cl[i-1].last_message_id)
        ufds[i].events |= POLLOUT;
      if (ufds[i].revents & POLLOUT) { // lze odeslat
        printf("A nco mu poleme.\n");
        if (packetstate(&cl[i-1].outpacket) == P_EMPTY)
          // a te najdeme prvn zprvu, kterou klient neml 
          for (j = 0; j < message_id; j++)
            if (j > cl[i-1].last_message_id) {
              printf("Zprvu %d\n", j);
              wrappacket(&cl[i-1].outpacket, ml[j].length, ml[j].message);
              break;
            }
        printf("Poslme klientovi [%s] zprvu.\n", cl[i-1].nick);
        if (send_packet(cl[i-1].fd, &cl[i-1].outpacket) == 1) { // odeleme
          clearpacket(&cl[i-1].outpacket);
          printf("Poslno\n");
          for (j = 0; j < message_id; j++)
            if (j > cl[i-1].last_message_id)
              cl[i-1].last_message_id = j;
          ufds[i].events &= ~POLLOUT;
        }
      }
    }
  }

  freepacket(&inpacket);
  
  close(s);
  return 0;
}
