#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include "protocol.h"

// penosov protokol

#ifdef IRIX_MODE
// debiln, e? :-)
#define MSG_NOSIGNAL 0
#endif

// praktitj funkce pro posln dat
ssize_t write_all(int fd, void *data, size_t length) {
  ssize_t n, n2 = 0;
  while (n2 < length) {
    n = send(fd, (void*)((size_t)data+n2), length-n2, MSG_NOSIGNAL);
    if (n < 0) return n;
    if (n == 0) return n2;
    n2 += n;
    length -= n;
  }
  return n2;
}

// peruovateln odesla paket
int send_packet(int fd, struct chat_packet * cp) {
  ssize_t n;
  if (cp->length == 0) {
#ifdef DEBUG    
    printf("Nothing to send!\n");
#endif    
    return 1;
  }
  if (cp->next == 0) { // poprv
    unsigned long int l = htonl(cp->length);
#ifdef DEBUG    
    printf("Send %d: ", cp->length);
#endif    
    n = write_all(fd, &l, sizeof(unsigned long int));
#ifdef DEBUG
    printf("header - %d bytes sent\n", n);
#endif    
    if (n < 0) return -1;
  }
#ifdef DEBUG
  printf("Sending %d - %d bytes\n", cp->next, cp->length);
#endif  
  n = send(fd, (void*)((size_t)cp->data + cp->next), cp->length - cp->next,
    MSG_NOSIGNAL);
#ifdef DEBUG
  printf("Sent: %d\n", n);
#endif
  if (n < 0) return n;
  cp->next += n;
  return (cp->next == cp->length);
}

// peruovateln pijma paket
int recv_packet(int fd, struct chat_packet * cp) {
  ssize_t n;
  if (cp->next == 0) { // poprv
    unsigned long int l;
    n = recv(fd, &l, sizeof(unsigned long int), MSG_PEEK | MSG_WAITALL);
#ifdef DEBUG    
    printf("Recv test: %d bytes; %lu length\n", n, ntohl(l));
#endif    
    if (n < 0) return -1;
    if (n == 0) { errno = EPIPE; return -1; }
    if (ntohl(l) > cp->size) { errno = EFBIG; return -2; } // kontrola rozsahu
    n = recv(fd, &l, sizeof(unsigned long int), MSG_WAITALL); // znovu
#ifdef DEBUG    
    printf("Recv real: %d; %lu\n", n, ntohl(l));
#endif
    if (n < 0) return -1;
    if (n == 0) { errno = EPIPE; return -1; }
    cp->length = ntohl(l);
  }
  if (cp->next == cp->length) return 1;
#ifdef DEBUG
  printf("Receiving: %d - %d bytes", cp->next, cp->length);
#endif  
  n = recv(fd, (void*)((size_t)cp->data + cp->next), cp->length - cp->next, 0);
#ifdef DEBUG
  printf("Received: %d\n", n);
#endif
  if (n < 0) return n;
  if (n == 0) { // petrhl se spoj?
    errno = EPIPE;
    return -1;
  }
  cp->next += n;
  return (cp->next == cp->length);
}

// vytvoen paketu
int createpacket(struct chat_packet *cp, int size) {
  if ((cp->data = malloc(size)) == NULL) {
    errno = ENOMEM;
    return -1;
  }
  cp->length = cp->next = 0;
  cp->size = size;
  return 0;
}

// odstrann paketu
void freepacket(struct chat_packet *cp) {
  if (cp->size != 0 && cp->data != NULL) free(cp->data);
}

// zabalen dat do paketu
void wrappacket(struct chat_packet *cp, unsigned long int length, void *data) {
  cp->length = length;
  cp->size = 0; // velikost alokovanch dat je nulov
  cp->next = 0;
  cp->data = data;
}

// stav paketu
int packetstate(const struct chat_packet *cp) {
  if (cp == NULL) return P_VOID;
  if (cp->next == 0) return P_EMPTY;
  if (cp->next == cp->size) return P_FINISHED;
  return P_BUSY;
}

// vyitn paketu
void clearpacket(struct chat_packet *cp) {
  cp->length = cp->next = 0;
}
